import java.util.Comparator;
import java.util.Random;
import java.util.TreeSet;

// This agent was written by Chandramohan Sol and Erik Hatfield
class HatfieldErik implements IAgent
{
  int iter, dodgeIter;
  int index, atk, def, kill; // a temporary value used to pass values around
  float lowestCost;
  Random Randomizer;


  HatfieldErik() {
    reset();
    Randomizer = new Random(1234);
    lowestCost = 99999;
  }

  public void reset() {
    iter = 0;
    lowestCost = 99999;
  }

  public static float sq_dist(float x1, float y1, float x2, float y2) {
    return (x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2);
  }
  public static float getDistance (float x1, float y1, float x2, float y2){
    return (float)Math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2)*(y1 - y2));
  }

  static class Shadow implements IAgent
  {
    float dx;
    float dy;

    Shadow(float destX, float destY) {
      dx = destX;
      dy = destY;
    }

    public void reset() {

    }

    public void update(Model m) {
      for(int i = 0; i < 3; i++) {
        if (dx > 0 && dx < Model.XMAX && dy > 0 && dy < Model.YMAX)
          m.setDestination(i, dx, dy);
      }
    }
  }

  static class OpponentShadow implements IAgent
  {
    int index;
    OpponentShadow() {
    }

    public void reset() {
    }

    float nearestOpponent(Model m, float x, float y) {
      index = -1;
      float dd = Float.MAX_VALUE;
      for(int i = 0; i < m.getSpriteCountOpponent(); i++) {
        if(m.getEnergyOpponent(i) < 0)
          continue; // don't care about dead opponents
        float d = sq_dist(x, y, m.getXOpponent(i), m.getYOpponent(i));
        if(d < dd) {
          dd = d;
          index = i;
        }
      }
      return dd;
    }

    float nearestBombTarget(Model m, float x, float y) {
      index = -1;
      float dd = Float.MAX_VALUE;
      for(int i = 0; i < m.getBombCount(); i++) {
        float d = sq_dist(x, y, m.getBombTargetX(i), m.getBombTargetY(i));
        if(d < dd) {
          dd = d;
          index = i;
        }
      }
      return dd;
    }

    void avoidBombs(Model m, int i) {
      if(nearestBombTarget(m, m.getX(i), m.getY(i)) <= 2.0f * Model.BLAST_RADIUS * Model.BLAST_RADIUS) {
        float dx;
        dx = m.getX(i) + m.getBombTargetX(index);
        float dy;
        if (m.getY(i) > Model.YFLAG_OPPONENT)
          dy = m.getY(i) - m.getBombTargetY(index);
        else
          dy = m.getY(i) + m.getBombTargetY(index);
        if(dx == 0 && dy == 0)
          dx = 1.0f;
        m.setDestination(i, m.getX(i) + dx, m.getY(i) + dy);
      }
    }

    public void update(Model m) {
      for(int i = 0; i < m.getSpriteCountSelf(); i++) {
        nearestOpponent(m, m.getX(i), m.getY(i));

        if(index >= 0) {
          float enemyX = m.getXOpponent(index);
          float enemyY = m.getYOpponent(index);
          if(sq_dist(enemyX, enemyY, m.getX(i), m.getY(i)) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
            m.throwBomb(i, enemyX, enemyY);
        }
        avoidBombs(m, i);
      }
    }
  }


  float nearestBombTarget(Model m, float x, float y) {
    index = -1;
    float dd = Float.MAX_VALUE;
    for(int i = 0; i < m.getBombCount(); i++) {
      float d = sq_dist(x, y, m.getBombTargetX(i), m.getBombTargetY(i));
      if(d < dd) {
        dd = d;
        index = i;
      }
    }
    return dd;
  }

  float nearestOpponent(Model m, float x, float y) {
    index = -1;
    float dd = Float.MAX_VALUE;
    for(int i = 0; i < m.getSpriteCountOpponent(); i++) {
      if(m.getEnergyOpponent(i) < 0)
        continue; // don't care about dead opponents
      float d = sq_dist(x, y, m.getXOpponent(i), m.getYOpponent(i));
      if(d < dd) {
        dd = d;
        index = i;
      }
    }
    return dd;
  }

  float bestDefender(Model m) {
    def = -1;
    float dd = Float.MAX_VALUE;
    for(int i = 0; i < m.getSpriteCountSelf(); i++) {
      if(m.getEnergySelf(i) < 0)
        continue; // don't care about dead agents
      float d = sq_dist(Model.XFLAG, Model.YFLAG, m.getX(i), m.getY(i));
      if(d < dd) {
        dd = d;
        def = i;
      }
    }
    return dd;
  }

  float bestAttacker(Model m) {
    atk = -1;
    float dd = Float.MAX_VALUE;
    for(int i = 0; i < m.getSpriteCountSelf(); i++) {
      if(m.getEnergySelf(i) < 0)
        continue; // don't care about dead agents
      float d = sq_dist(Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT, m.getX(i), m.getY(i));
      if(d < dd) {
        dd = d;
        atk = i;
      }
    }
    return dd;
  }

  float nearestOpponentToFlag(Model m) {
    index = -1;
    float dd = Float.MAX_VALUE;
    for(int i = 0; i < m.getSpriteCountOpponent(); i++) {
      if(m.getEnergyOpponent(i) < 0)
        continue; // don't care about dead opponents
      float d = sq_dist(Model.XFLAG, Model.YFLAG, m.getXOpponent(i), m.getYOpponent(i));
      if(d < dd) {
        dd = d;
        index = i;
      }
    }
    return dd;
  }

  float nextToRevive(Model m, float x, float y) {
    index = -1;
    float dd = Float.MAX_VALUE;
    for(int i = 0; i < m.getSpriteCountOpponent(); i++) {
      if(m.getEnergyOpponent(i) > 0)
        continue; // consider only dead opponents
      float d = sq_dist(Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT, m.getXOpponent(i), m.getYOpponent(i));
      if(d < dd) {
        dd = d;
        index = i;
      }
    }
    return dd;
  }

  float distanceToOpponentFlag(Model m, int i){
    float d = sq_dist(Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT, m.getXOpponent(i), m.getYOpponent(i));
    return d;
  }

  void avoidBombs(Model m, int i) {
    if(nearestBombTarget(m, m.getX(i), m.getY(i)) <= 2.0f * Model.BLAST_RADIUS * Model.BLAST_RADIUS) {
      if (i == def && Model.XFLAG > m.getX(i) && m.getEnergySelf(i) >= .5){
        float dx;
        dx = m.getX(i) + m.getBombTargetX(index);
        float dy;
        if (m.getY(i) > Model.YFLAG)
          dy = m.getY(i) + (m.getBombTargetY(index)+5);
        else
          dy = m.getY(i) - (m.getBombTargetY(index)+5);
        if(dx == 0 && dy == 0)
          dx = 1.0f;
        m.setDestination(i, m.getX(i) + dx, m.getY(i) + dy);
      }
      else
        forkAvoidBombs(m, i, m.getBombTargetX(index), m.getBombTargetY(index));
//			else{
//				float dx;
//				if (i == def)
//					dx = m.getX(i) - (m.getBombTargetX(index));
//				else
//					dx = m.getX(i) + m.getBombTargetX(index);
//
//				float dy;
//				if (i == def)
//					if (m.getY(i) > Model.YFLAG)
//						 dy = m.getY(i) + (m.getBombTargetY(index)+5);
//					else
//						dy = m.getY(i) - (m.getBombTargetY(index)+5);
//				else
//					if (m.getY(i) > Model.YFLAG_OPPONENT)
//						 dy = m.getY(i) - m.getBombTargetY(index);
//					else
//						dy = m.getY(i) + m.getBombTargetY(index);
//
//				if(dx == 0 && dy == 0)
//					dx = 1.0f;
//				m.setDestination(i, m.getX(i) + dx, m.getY(i) + dy);
//			}
      dodgeIter++;
    }
  }

  boolean isSafe(Model m, int i, float buffer){
    float myX = m.getX(i);
    float myY = m.getY(i);
    return (nearestOpponent(m, myX, myY) > (((Model.MAX_THROW_RADIUS + Model.BLAST_RADIUS) * (Model.MAX_THROW_RADIUS + Model.BLAST_RADIUS))+buffer));
  }

  int numOpponentsInRange(Model m, int i, float buffer){
    float myX = m.getX(i);
    float myY = m.getY(i);
    int numInRange = 0;
    for(int j = 0; j < m.getSpriteCountOpponent(); j++)
      if(m.getEnergyOpponent(j) > 0 && sq_dist(myX, myY, m.getXOpponent(j), m.getYOpponent(j)) <= (((Model.MAX_THROW_RADIUS + Model.BLAST_RADIUS) * (Model.MAX_THROW_RADIUS + Model.BLAST_RADIUS)))+buffer)
        numInRange++;
    return numInRange;
  }

  int numOpponentsAlive(Model m){
    int numAlive = 0;
    for(int i = 0; i < m.getSpriteCountOpponent(); i++)
      if(m.getEnergyOpponent(i) > 0)
        numAlive++;
    return numAlive;
  }

  int numFriendsAlive(Model m){
    int numAlive = 0;
    for(int i = 0; i < m.getSpriteCountOpponent(); i++)
      if(m.getEnergySelf(i) > 0)
        numAlive++;
    return numAlive;
  }

  float enemyDistToFlag(Model m){
    float closestDist = Float.MAX_VALUE;
    float dist;
    for(int i = 0; i < 3; i++){
      float eX = m.getXOpponent(i);
      float eY = m.getYOpponent(i);

      dist = getDistance(Model.XFLAG, Model.YFLAG, eX, eY);
      if(dist < closestDist){
        closestDist = dist;
      }
    }
    return closestDist;
  }

  boolean revivalImminent(Model m, float offset){
    for(int i = 0; i < m.getSpriteCountOpponent(); i++)
      if(m.getEnergyOpponent(i) <= 0)
        if (sq_dist(Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT, m.getXOpponent(i), m.getYOpponent(i)) < offset*offset)
          return true;
    return false;
  }


  void forkAvoidBombs(Model m, int i, float bombX, float bombY) {
    dodgeIter++;

    float bestX = m.getX(i);
    float bestY = m.getY(i);
//		float best_iEnergy = Model.XMAX*Model.XMAX;
    float bestDodge = 0;
    for(int sim = 0; sim < 8; sim++) { // try 8 candidate destinations
      float x = (float)(m.getX(i)+(Math.cos((Math.PI*sim)/4)*Model.BLAST_RADIUS));
      float y = (float)(m.getY(i)+(Math.sin((Math.PI*sim)/4)*Model.BLAST_RADIUS));

      // Fork the universe and simulate it for 10 time-steps
      Controller cFork = m.getController().fork(new Shadow(x, y), new OpponentShadow());
      Model mFork = cFork.getModel();
      for(int j = 0; j < 10; j++)
        cFork.update();

      // See how close the current sprite got to the opponent's flag in the forked universe
//			float sqd = sq_dist(mFork.getX(i), mFork.getY(i), XGoal, YGoal);
//			float iEnergy = sq_dist(mFork.getX(i), mFork.getY(i), Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
      float dodge = sq_dist(mFork.getX(i), mFork.getY(i), bombX, bombY);

      if(dodge > bestDodge) {
        bestDodge = dodge;
        bestX = x;
        bestY = y;
      }
    }

    // Head for the point that worked out best in simulation
    m.setDestination(i, bestX, bestY);
  }

  void pickPath(Model m, int i, float XGoal, float YGoal) {

    float bestX = m.getX(i);
    float bestY = m.getY(i);
//		float best_sqdist = sq_dist(bestX, bestY, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
    float best_sqdist = 0;

    float startX = m.getX(i);
    float startY = m.getY(i);

    for(int sim = 0; sim < 8; sim++) { // try 8 candidate destinations
      float x = (float)(m.getX(i)+(Math.cos((Math.PI*sim)/4)*100));
      float y = (float)(m.getY(i)+(Math.sin((Math.PI*sim)/4)*100));

      // Fork the universe and simulate it for 10 time-steps
      Controller cFork = m.getController().fork(new Shadow(x, y), new OpponentShadow());
      Model mFork = cFork.getModel();
      for(int j = 0; j < 3; j++)
        cFork.update();
//			float covered = sq_dist(mFork.getX(i), mFork.getY(i), startX, startY);


      // See how much ground was covered in the forked universe
      float sqd = sq_dist(mFork.getX(i), mFork.getY(i), startX, startY);;
      if(sqd > best_sqdist) {
        best_sqdist = sqd;
        bestX = x;
        bestY = y;
      }
    }

    // Head for the point that worked out best in simulation
    m.setDestination(i, bestX, bestY);

    // Shoot at the flag if I can hit it
    if(sq_dist(m.getX(i), m.getY(i), Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
      m.throwBomb(i, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
  }

  void beDefender(Model m, int i) {
    PathFinder pathFinder = new PathFinder();
    pathFinder.lowestCost = lowestCost;
    GameState path;
    // Find the opponent nearest to my flag
    nearestOpponentToFlag(m);
    if(index >= 0) {
      float enemyX = m.getXOpponent(index);
      float enemyY = m.getYOpponent(index);

      // Throw bombs if the enemy gets close enough
      if(sq_dist(enemyX, enemyY, m.getX(i), m.getY(i)) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
        m.throwBomb(i, enemyX, enemyY);

      // Stay between the enemy and my flag
      if(sq_dist(enemyX, enemyY, Model.XFLAG, Model.YFLAG) > Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
        m.setDestination(i, 0.5f * (Model.XFLAG + enemyX), 0.5f * (Model.YFLAG + enemyY));
      else
        m.setDestination(i, 0.75f * (Model.XFLAG + enemyX), 0.5f * (Model.YFLAG + enemyY));

    }
    else {
      // Guard the flag
      path = pathFinder.bestPath(m, i, Model.XFLAG + Model.MAX_THROW_RADIUS, Model.YFLAG);
      m.setDestination(i, Model.XFLAG + Model.MAX_THROW_RADIUS, Model.YFLAG);
    }

    // If I don't have enough energy to throw a bomb, rest
    if(m.getEnergySelf(i) < Model.BOMB_COST)
      m.setDestination(i, m.getX(i), m.getY(i));

    // Try not to die
    avoidBombs(m, i);
  }

  void beFlagAttacker(Model m, int i) {
    // Head for the opponent's flag
    PathFinder pathFinder = new PathFinder();
    pathFinder.lowestCost = lowestCost;
    GameState path;

    path = pathFinder.bestPath(m, i, Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS + 1, Model.YFLAG_OPPONENT);
    if(path != null){
      move(m, i, path.x, path.y);
    }
    else {
      m.setDestination(i, Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS + 1, Model.YFLAG_OPPONENT);
    }


    float myX = m.getX(i);
    float myY = m.getY(i);

    if (isSafe(m, i, 20)){
      if (m.getEnergySelf(i) < 1 && numOpponentsAlive(m) > 0 && nextToRevive(m, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT)/Model.BROKEN_CRAWL_RATE > sq_dist(myX, myY, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT)*.5)
        m.setDestination(i, myX, myY); // Rest
      else{
//				if(iter % 10 == 0)
        pickPath(m, i, Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS + 1, Model.YFLAG_OPPONENT);
//				else
        // Avoid opponents
        nearestOpponent(m, myX, myY);

        if(index >= 0) {
          float enemyX = m.getXOpponent(index);
          float enemyY = m.getYOpponent(index);
          if(sq_dist(enemyX, enemyY, myX, myY) <= (Model.MAX_THROW_RADIUS + Model.BLAST_RADIUS*1.5) * (Model.MAX_THROW_RADIUS + Model.BLAST_RADIUS*1.5))
//						if (iter%1==0)
            pickPath(m, i, myX + 15.0f * (myX - enemyX), myY + 15.0f * (myY - enemyY));
//						else
//							m.setDestination(i, myX + 15.0f * (myX - enemyX), myY + 15.0f * (myY - enemyY));
        }

        if (numOpponentsInRange(m, i, 0) == 0 && !revivalImminent(m, 100)){
//					if (iter%20==0)
//						pickPath(m, i, Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS + 1, Model.YFLAG_OPPONENT);
//					else
          m.setDestination(i, Model.XFLAG_OPPONENT - 100, Model.YFLAG_OPPONENT);
        }
        if (revivalImminent(m, 100)){
          m.setDestination(i, Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS +1, Model.YFLAG_OPPONENT);
          float d = nextToRevive(m, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
          if (index >= 0){
            if (d <= 100){
              float enemyX = m.getXOpponent(index);
              float enemyY = m.getYOpponent(index);
              if(sq_dist(enemyX, enemyY, m.getX(i), m.getY(i)) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
                m.throwBomb(i, enemyX, enemyY);
            }
            else
            if(sq_dist(m.getX(i), m.getY(i), Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
              m.throwBomb(i, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
          }
        }
        else{
          if (iter%35==0)
            pickPath(m, i, Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS + 1, Model.YFLAG_OPPONENT);
          else
            m.setDestination(i, Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS + 1, Model.YFLAG_OPPONENT);

          // Shoot at the flag if I can hit it
          if(sq_dist(m.getX(i), m.getY(i), Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
            m.throwBomb(i, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
        }
      }
    }
    else{ //it is not safe!
      float distance = nearestOpponent(m, myX, myY);

      if(index >= 0) {
        float enemyX = m.getXOpponent(index);
        float enemyY = m.getYOpponent(index);

        // If the opponent is close enough to shoot at me...
//				if ((sq_dist(myX, myY, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) < sq_dist(myX, myY, enemyX, enemyY)) && ((myX>Model.XFLAG_OPPONENT && enemyX<Model.XFLAG_OPPONENT)||(myX<Model.XFLAG_OPPONENT && enemyX>Model.XFLAG_OPPONENT))){
////					m.setDestination(i, Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS + 1, Model.YFLAG_OPPONENT);
        if((sq_dist(m.getX(i), m.getY(i), Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS) && (sq_dist(enemyX, enemyY, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) < Model.BLAST_RADIUS*Model.BLAST_RADIUS)){
          m.throwBomb(i, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
        }
        else{
          if (distance > Model.BLAST_RADIUS || m.getEnergyOpponent(index)-Model.BOMB_DAMAGE_TO_SPRITE <=0 || m.getEnergySelf(i)-(Model.BOMB_COST+Model.BOMB_DAMAGE_TO_SPRITE) > 0)
            if (nearestBombTarget(m, enemyX, enemyY) > (Model.BLAST_RADIUS*Model.BLAST_RADIUS)+Model.BLAST_RADIUS+1f)
              m.throwBomb(i, enemyX, enemyY);
//					m.setDestination(i, myX + 10.0f * (myX - enemyX), myY + 10.0f * (myY - enemyY)); // Flee
        }
      }
    }

//		// Shoot at the flag if I can hit it
//		if(sq_dist(m.getX(i), m.getY(i), Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS) {
//			m.throwBomb(i, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
//		}
//		else{
//			nearestOpponent(m, myX, myY);
//			if(index >= 0) {
//				float enemyX = m.getXOpponent(index);
//				float enemyY = m.getYOpponent(index);
//
//				if(m.getEnergySelf(i) >= m.getEnergyOpponent(index)) {
//
//					// Get close enough to throw a bomb at the enemy
//					float dx = myX - enemyX;
//					float dy = myY - enemyY;
//					float t = 1.0f / Math.max(Model.EPSILON, (float)Math.sqrt(dx * dx + dy * dy));
//					dx *= t;
//					dy *= t;
//					m.setDestination(i, enemyX + dx * (Model.MAX_THROW_RADIUS - Model.EPSILON), enemyY + dy * (Model.MAX_THROW_RADIUS - Model.EPSILON));
//
//					// Throw bombs
//					if(sq_dist(enemyX, enemyY, m.getX(i), m.getY(i)) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
//						m.throwBomb(i, enemyX, enemyY);
//				}
//				else{
//					m.setDestination(i, Model.XFLAG, Model.YFLAG);
//					atk = def;
//					def = i;
//				}
//
//			}
//		}


    // Try not to die

    nearestOpponent(m, myX, myY);
    if(index >= 0) {
      float enemyX = m.getXOpponent(index);
      float enemyY = m.getYOpponent(index);
      if (!(m.getEnergySelf(i)>=.2 && m.getEnergySelf(i)-(Model.BOMB_DAMAGE_TO_SPRITE+Model.BOMB_COST) > m.getEnergyOpponent(index)))
        avoidBombs(m, i);

    }
  }

  void beAggressor(Model m, int i) {
    float myX = m.getX(i);
    float myY = m.getY(i);
    boolean rush = false;

    // Find the opponent nearest to me
//		nearestOpponent(m, myX, myY);
    nearestOpponentToFlag(m);
    GameState path = new GameState(null);
    PathFinder pathFinder = new PathFinder();
    if(index >= 0) {
      float enemyX = m.getXOpponent(index);
      float enemyY = m.getYOpponent(index);

      if(m.getEnergySelf(i) >= m.getEnergyOpponent(index)) {
        // Get close enough to throw a bomb at the enemy
        float dx = myX - enemyX;
        float dy = myY - enemyY;
        float t = 1.0f / Math.max(Model.EPSILON, (float)Math.sqrt(dx * dx + dy * dy));
        dx *= t;
        dy *= t;
        if (iter%5==0) {
          pickPath(m, i, enemyX + dx * (Model.MAX_THROW_RADIUS - Model.EPSILON), enemyY + dy * (Model.MAX_THROW_RADIUS - Model.EPSILON));
        }
        else {
          m.setDestination(i, enemyX + dx * (Model.MAX_THROW_RADIUS - Model.EPSILON), enemyY + dy * (Model.MAX_THROW_RADIUS - Model.EPSILON));
        }
        // Throw bombs
        if(sq_dist(enemyX, enemyY, m.getX(i), m.getY(i)) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
          m.throwBomb(i, enemyX, enemyY);
        if(m.getEnergyOpponent(index) < .3){
          rush = true;
        }
      }
      else {

        // If the opponent is close enough to shoot at me...
        if(sq_dist(enemyX, enemyY, myX, myY) <= (Model.MAX_THROW_RADIUS + Model.BLAST_RADIUS) * (Model.MAX_THROW_RADIUS + Model.BLAST_RADIUS) + 5.0f) {
//					if (iter%5==0)
          pickPath(m, i, myX + 10.0f * (myX - enemyX), myY + 10.0f * (myY - enemyY));
//					else
//						m.setDestination(i, myX + 5.0f * (myX - enemyX), myY + 5.0f * (myY - enemyY)); // Flee
        }
        else {
          m.setDestination(i, myX, myY); // Rest
        }
      }
    }
    else {
      // Head for the opponent's flag
      pathFinder.xDest = (int) (Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS)/10 * 10;
      pathFinder.yDest = (int) (Model.YFLAG_OPPONENT - Model.MAX_THROW_RADIUS) /10 * 10;
      path = pathFinder.bestPath(m);
      //path = pathFinder.bestPath(m);
      m.setDestination(i, Model.XFLAG_OPPONENT - Model.MAX_THROW_RADIUS + 1, Model.YFLAG_OPPONENT);

      // Shoot at the flag if I can hit it
      if(sq_dist(m.getX(i), m.getY(i), Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS) {
        m.throwBomb(i, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
      }
    }
    if(!rush){
      // Try not to die
      avoidBombs(m, i);
    }
  }

  void beRush(Model m, int i){

  }

  void chooseRole(Model m, int i){
    if (iter%30 == 0){
      bestAttacker(m);
      bestDefender(m);
    }
    while (true){
      if (((numOpponentsAlive(m) <= 1 && nextToRevive(m, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) >= 100*100) || (numOpponentsAlive(m) == 0)) && nearestOpponentToFlag(m) >= 500*500){
//				if (index >= 0){
//					if (def == i)
//						if (m.getEnergySelf(i) - Model.BOMB_DAMAGE_TO_SPRITE <=0 && m.getEnergyOpponent(index) - Model.BOMB_DAMAGE_TO_SPRITE > m.getEnergySelf(i))
//							beDefender(m, i);
//						else
//							beAggressor(m, i);
//					else
//						beFlagAttacker(m, i);
//				}
//				else
        beFlagAttacker(m, i);
        break;
      }
      else{
        if (def == i){
          if(numOpponentsAlive(m) < numFriendsAlive(m) && enemyDistToFlag(m) > 1500)
            beFlagAttacker(m, i);
          else
            beDefender(m, i);
          break;
        }
        if (atk == i){
          if (numOpponentsInRange(m, def, 250) <= 1 || m.getEnergySelf(def)>.75)
            beFlagAttacker(m, i);
          else
            beAggressor(m, i);
          break;
        }
        if (m.getFlagEnergyOpponent() < .5 && numOpponentsInRange(m, def, 250) == 0)
          beFlagAttacker(m, i);
        else if(numOpponentsAlive(m) >= numFriendsAlive(m) && enemyDistToFlag(m) < 750)
          beDefender(m, i);
        else
          beAggressor(m, i);
        break;
      }

    }
  }

  void move(Model m, int i, float xDest, float yDest){
    float curX = m.getX(i);
    float curY = m.getY(i);

    m.setDestination(i, curX + xDest, curY + yDest);
  }

  public void update(Model m) {
    if(lowestCost == 99999){
      lowestCost = getLowestCost(m);
    }
    for(int i = 0; i < m.getSpriteCountSelf(); i++) {
      long time = m.getTimeBalance();
      if(m.getEnergySelf(i) > 0 && iter > 200) // && iter > 545
        chooseRole(m, i);
      if(time < 1000){
        System.out.println("OUT OF TIME");
      }
    }
//		beFlagAttacker(m, 0);
//		beAggressor(m, 1);
//		beDefender(m, 2);

    iter++;
  }
  public float getLowestCost(Model m){
    float lowest = -9999, compareValue = 0;
    for(int x = 0; x < 1200; x += 10){
      for(int y = 0; y < 600; y += 10){
        compareValue = m.getTravelSpeed(x, y);
        if(compareValue > lowest){
          lowest = compareValue;
        }
      }
    }
    return lowest;
  }

  static class PathFinder{
    int spriteNumber;
    float xDest, yDest;
    float lowestCost;
    TreeSet<GameState> frontier;
    TreeSet<GameState> stateTracker;

    PathFinder(){
      spriteNumber = 0;
      xDest = 0;
      yDest = 0;
      lowestCost = 0f;
      frontier = new TreeSet<>(new CostComparator());
      stateTracker = new TreeSet<>(new StateComparator());
    }

    PathFinder(int spriteNumber, int xDest, int yDest, int lowestCost){
      this.spriteNumber = spriteNumber;
      this.xDest = xDest;
      this.yDest = yDest;
      this.lowestCost = lowestCost;
    }

    GameState bestPath(Model m){
      //System.out.println("bestPath");
      GameState finalState = new GameState(null);

      int curX, curY;

      curX = (int) m.getX(spriteNumber)/10*10;
      curY = (int) m.getY(spriteNumber)/10*10;

      curX = curX / 10 * 10;
      curY = curY / 10 * 10;

      if(curX != xDest || curY != yDest) {
        GameState initialState = new GameState(null);
        GameState checkState;
        initialState.costRemaining = getHeuristic(initialState, m);
        frontier.add(initialState);
        stateTracker.add(initialState);

        while (!frontier.isEmpty()) {
          checkState = frontier.pollFirst();

          if(checkState.x + curX == xDest && checkState.y + curY == yDest){
            finalState = copyState(checkState);
            break;
          }

          if(checkValid(checkState, 10, 0, m)){
            addChild(checkState, 10, 0, m);
          }

          // 10 10
          if(checkValid(checkState, 10, 10, m)){
            addChild(checkState, 10, 10, m);
          }

          // 10 -10
          if(checkValid(checkState, 10, -10, m)){
            addChild(checkState, 10, -10, m);
          }

          // 0 10
          if(checkValid(checkState, 0, 10, m)){
            addChild(checkState, 0, 10, m);
          }

          // 0 -10
          if(checkValid(checkState, 0, -10, m)){
            addChild(checkState, 0, -10, m);
          }

          // -10 0
          if(checkValid(checkState, -10, 0, m)){
            addChild(checkState, -10, 0, m);
          }

          // -10 10
          if(checkValid(checkState, -10, 10, m)){
            addChild(checkState, -10, 10, m);
          }

          // -10 -10
          if(checkValid(checkState, -10, -10, m)){
            addChild(checkState, -10, -10, m);
          }
          System.out.print("");
        }
      }
      if(finalState.parent == null){
        System.out.println("No Path Found");
      }
      frontier.clear();
      stateTracker.clear();
      return flipPath(finalState);
    }

    GameState bestPath(Model m, int spriteNumber, float xDest, float yDest){
      //System.out.println("bestPath");
      GameState finalState = new GameState(null);
      this.spriteNumber = spriteNumber;
      int curX, curY;

      curX = (int) m.getX(spriteNumber)/10*10;
      curY = (int) m.getY(spriteNumber)/10*10;

      curX = curX / 10 * 10;
      curY = curY / 10 * 10;

      this.xDest = (int) xDest/10*10;
      this.yDest = (int) yDest/10*10;

      if(curX != this.xDest || curY != this.yDest) {
        GameState initialState = new GameState(null);
        GameState checkState;
        initialState.costRemaining = getHeuristic(initialState, m);
        frontier.add(initialState);
        stateTracker.add(initialState);

        while (!frontier.isEmpty()) {
          checkState = frontier.pollFirst();

          if(checkState.x + curX == this.xDest && checkState.y + curY == this.yDest){
            finalState = copyState(checkState);
            break;
          }

          if(checkValid(checkState, 10, 0, m)){
            addChild(checkState, 10, 0, m);
          }

          // 10 10
          if(checkValid(checkState,10, 10, m)){
            addChild(checkState, 10, 10, m);
          }

          // 10 -10
          if(checkValid(checkState, 10, -10, m)){
            addChild(checkState, 10, -10, m);
          }

          // 0 10
          if(checkValid(checkState, 0, 10, m)){
            addChild(checkState, 0, 10, m);
          }

          // 0 -10
          if(checkValid(checkState, 0, -10, m)){
            addChild(checkState, 0, -10, m);
          }

          // -10 0
          if(checkValid(checkState, -10, 0, m)){
            addChild(checkState, -10, 0, m);
          }

          // -10 10
          if(checkValid(checkState, -10, 10, m)){
            addChild(checkState, -10, 10, m);
          }

          // -10 -10
          if(checkValid(checkState, -10, -10, m)){
            addChild(checkState, -10, -10, m);
          }
          System.out.print("");
        }
      }
      if(finalState.parent == null){
        //System.out.println("no escape path found");
      }
      frontier.clear();
      stateTracker.clear();
      return flipPath(finalState);
    }

    GameState flipPath(GameState pathToFlipp){
      GameState flippedPath = new GameState(null);
      GameState pathToFlip = new GameState(pathToFlipp);
      //System.out.println("path flip");
      pathToFlip = pathToFlip.parent;
      while(pathToFlip.parent != null){
        if(pathToFlip.distanceRemaining == 0){
          //System.out.println("OOPS");
        }
        flippedPath.distanceRemaining = pathToFlip.distanceRemaining;
        flippedPath.costRemaining = pathToFlip.costRemaining;
        flippedPath.cost = pathToFlip.cost;
        flippedPath.x = pathToFlip.x;
        flippedPath.y = pathToFlip.y;

        flippedPath = new GameState(flippedPath);
        pathToFlip = pathToFlip.parent;
      }
      //System.out.println("Finished path flip");
      if(flippedPath.parent == null){
        //System.out.println("no flipped path");
      }
      return flippedPath.parent;
    }

    private double getHeuristic(GameState startState, Model m){
      double heuristic = 0;
      double distance = 0;
      float x1, x2, y1, y2;

      x1 = (int) (startState.x + m.getX(spriteNumber) / 10 * 10);
      y1 = (int) (startState.y + m.getY(spriteNumber) / 10 * 10);

      x2 = xDest;
      y2 = yDest;

      distance = getDistance(x1, y1, x2, y2);
      heuristic = distance / lowestCost;

      return heuristic;
    }
    private double getHeuristic(GameState startState, Model m, float xDest, float yDest){
      double heuristic = 0;
      double distance = 0;
      float x1, x2, y1, y2;

      x1 = (int) (startState.x + m.getX(spriteNumber) / 10 * 10);
      y1 = (int) (startState.y + m.getY(spriteNumber) / 10 * 10);

      x2 = xDest;
      y2 = yDest;
      distance = getDistance(x1, y1, x2, y2);

      heuristic = distance / lowestCost;

      return heuristic;
    }

    GameState copyState(GameState stateToCopy){
      GameState newState = new GameState(stateToCopy.parent);
      newState.x = stateToCopy.x;
      newState.y = stateToCopy.y;
      newState.cost = stateToCopy.cost;
      newState.costRemaining = stateToCopy.costRemaining;

      return newState;
    }

    private boolean checkValid(GameState startState, int x, int y, Model m) {
      if(startState.x + x + m.getX(spriteNumber) > 0 && startState.x + x + m.getX(spriteNumber) < 1200){
        if(startState.y + y + m.getY(spriteNumber) > 0 && startState.y + y + m.getY(spriteNumber) < 600){
          return true;
        }
      }
      return false;
    }
    private boolean checkValid(GameState startState, int spriteNumber, int x, int y, Model m) {
      if(startState.x + x + m.getX(spriteNumber) > 0 && startState.x + x + m.getX(spriteNumber) < 1200){
        if(startState.y + y + m.getY(spriteNumber) > 0 && startState.y + y + m.getY(spriteNumber) < 600){
          return true;
        }
      }
      return false;
    }

    private double getCost(GameState startState, int x, int y, Model m) {
      double distance;
      double speed;
      double cost;

      distance = Math.sqrt((x*x)+(y*y));
      speed = m.getTravelSpeed(m.getX(spriteNumber)+ x + startState.x, m.getY(spriteNumber)+y+startState.y);

      cost = distance/speed;

      return cost;
    }

    private double getCost(GameState startState, int spriteNumber, int x, int y, Model m) {
      double distance;
      double speed;
      double cost;

      distance = Math.sqrt((x*x)+(y*y));
      speed = m.getTravelSpeed(m.getX(spriteNumber)+ x + startState.x, m.getY(spriteNumber)+y+startState.y);

      cost = distance/speed;

      return cost;
    }

    private void addChild(GameState startState, int x, int y, Model m) {
      GameState child = new GameState(null);
      GameState oldChild;
      double cost, hueristic;
      //System.out.println("add child");
      cost = getCost(startState, x, y, m);
//			if(cost > 50){
//				cost *= 1000;
//			}
      hueristic = getHeuristic(startState, m);

      child.costRemaining = hueristic;
      child.x = startState.x + x;
      child.y = startState.y + y;
      child.distanceRemaining = getDistance(child.x, child.y, xDest, yDest);

      if(stateTracker.contains(child)){
        oldChild = stateTracker.floor(child);
        if(startState.cost + cost + hueristic < oldChild.cost + oldChild.costRemaining){
          oldChild.cost = startState.cost + cost;
          oldChild.costRemaining = hueristic;
          oldChild.distanceRemaining = child.distanceRemaining;
          oldChild.parent = startState;
        }
      } else {
        child.cost = startState.cost + cost;
        child.costRemaining = hueristic;
        child.parent = copyState(startState);
        while(frontier.contains(child)){
          child.cost -=.00001;
        }
        frontier.add(child);
        stateTracker.add(child);
      }
    }

  }
  static class GameState {
    public double cost;
    GameState parent;
    int x;
    int y;
    double costRemaining;
    float distanceRemaining;

    GameState(GameState prev){
      parent = prev;
      cost = 0;
      x = 0;
      y = 0;
      costRemaining = 0;
      distanceRemaining = 0f;
    }
    public void replaceParent(GameState prev){
      parent = prev;
      cost = prev.cost;
      x = prev.x;
      y = prev.y;
      costRemaining = prev.costRemaining;
      distanceRemaining = prev.distanceRemaining;

    }
  }
  static class StateComparator implements Comparator<GameState> {
    public int compare(GameState a, GameState b) {
      if (a.x < b.x)
        return -1;
      else if (a.x > b.x)
        return 1;
      else if (a.y < b.y)
        return -1;
      else if (a.y > b.y)
        return 1;
      return 0;
    }
  }
  static class CostComparator implements Comparator<GameState> {
    public int compare(GameState a, GameState b)
    {
      if(a.cost + a.costRemaining < b.cost + b.costRemaining)
        return -1;
      else if(a.cost + a.costRemaining > b.cost + b.costRemaining)
        return 1;
      return 0;
    }
  }

}


