import java.util.*;
import java.io.*;
import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.Color;
import java.lang.Math;

class PriestGabriel implements IAgent
{
	int iter;
	int index;
	PathPlanner planner;
	LinkedList<GameState> path;
	GameState startState;
	GameState goalState;
	GameState pather;
	GameState step;
	


	PriestGabriel()
	{
		path = new LinkedList<GameState>();
		reset();
	}

	public void reset()
	{
		iter = 0;
	}
	public static float sq_dist(float x1, float y1, float x2, float y2)
	{
		return (x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2);
	}

	public static float dist(float x_1, float y_1, float x_2, float y_2)
	{
		return (float)Math.sqrt((x_1 - x_2) * (x_1 - x_2) + (y_1 - y_2) * (y_1 - y_2));
	}

	float nearestBombTarget(Model m, float x, float y)
	{
			index = -1;
			float dd = Float.MAX_VALUE;
			for(int i = 0; i < m.getBombCount(); i++) {
				float d = sq_dist(x, y, m.getBombTargetX(i), m.getBombTargetY(i));
				if(d < dd) {
					dd = d;
					index = i;
				}
			}
			return dd;
	}

	float nearestOpponent(Model m, float x, float y) {
		index = -1;
		float dd = Float.MAX_VALUE;
		for(int i = 0; i < m.getSpriteCountOpponent(); i++) {
			if(m.getEnergyOpponent(i) < 0)
				continue; // don't care about dead opponents
			float d = sq_dist(x, y, m.getXOpponent(i), m.getYOpponent(i));
			if(d < dd) {
				dd = d;
				index = i;
			}
		}
		return dd;
	}

	void avoidBombs(Model m, int i) {
		if(nearestBombTarget(m, m.getX(i), m.getY(i)) <= 2.0f * Model.BLAST_RADIUS * Model.BLAST_RADIUS) {
			float dx = m.getX(i) - m.getBombTargetX(index);
			float dy = m.getY(i) - m.getBombTargetY(index);
			if(dx == 0 && dy == 0)
				dx = 1.0f;
			m.setDestination(i, m.getX(i) + dx * 10.0f, m.getY(i) + dy * 10.0f);
		}
	}
	void beDefender(Model m, int i) {
		// Find the opponent nearest to my flag
		nearestOpponent(m, Model.XFLAG, Model.YFLAG);
		if(index >= 0) {
			float enemyX = m.getXOpponent(index);
			float enemyY = m.getYOpponent(index);

			// Stay between the enemy and my flag
			m.setDestination(i, 0.5f * (Model.XFLAG + enemyX), 0.5f * (Model.YFLAG + enemyY));

			// Throw boms if the enemy gets close enough
			if(sq_dist(enemyX, enemyY, m.getX(i), m.getY(i)) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS)
				m.throwBomb(i, enemyX, enemyY);
		}
		else {
			// Guard the flag
			m.setDestination(i, Model.XFLAG + Model.MAX_THROW_RADIUS, Model.YFLAG);
		}

		// If I don't have enough energy to throw a bomb, rest
		if(m.getEnergySelf(i) < Model.BOMB_COST)
			m.setDestination(i, m.getX(i), m.getY(i));

		// Try not to die
		avoidBombs(m, i);
	}
	public void update(Model m) {

		// Find my player with the most energy.
		int leader = 0;
		if(m.getEnergySelf(1) > m.getEnergySelf(leader))
			leader = 1;
		if(m.getEnergySelf(2) > m.getEnergySelf(leader))
			leader = 2;

		// Find the enemy closest to the leader
		nearestOpponent(m, m.getX(leader), m.getY(leader));
		if(index >= 0) {
			beDefender(m, 0);
			beDefender(m, 1);
			beDefender(m, 2);
		}
		else {

			for(int i = 0; i < 3; i++) {

				// Head for the opponent's flag
				planner = new PathPlanner(m);
				GameState startState = new GameState(0.0f,null,m);
				GameState goalState = new GameState(0.0f,null,m);
				GameState pather = new GameState(0.0f,null,m);
				GameState step = new GameState(0.0f,null,m);
				path = new LinkedList<GameState>();
				startState.state[0] = Math.round(m.getX(i)/10)*10;
				startState.state[1] = Math.round(m.getY(i)/10)*10;

				goalState.state[0] = Math.round((Model.XFLAG_OPPONENT - 150) /10)*10;
				goalState.state[1] = Math.round(Model.YFLAG_OPPONENT/10)*10;
				if(!startState.stateEqualizer(goalState))
				{
					pather = planner.aStarSearch(startState,goalState, m);
					path = pather.queuePath();
					if(path.size() == 1)
					{
						m.setDestination(i, goalState.state[0], goalState.state[1]);
					}
					else
					{
						m.setDestination(i,path.get(1).state[0], path.get(1).state[1]);
					}

				}
				else
				{
				}
				// Shoot at the flag if I can hit it
				if(sq_dist(m.getX(i), m.getY(i), Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT) <= Model.MAX_THROW_RADIUS * Model.MAX_THROW_RADIUS) {
						m.throwBomb(i, Model.XFLAG_OPPONENT, Model.YFLAG_OPPONENT);
				}
			}
		}

		// Try not to die
		for(int i = 0; i < 3; i++)
				avoidBombs(m, i);
		iter++;
	}
	class GameState
{
	public float cost;
	GameState parent;
	float[] state;
	Model model;
	float heuristicA;

	GameState( float costs, GameState par, Model m)
	{
		model = m;
		state = new float[2];
		cost = costs;
		parent = par;
		heuristicA = bestValue();
		if(par == null)
		{
			state[0] = 0.0f;
			state[1] = 0.0f;
		}
		else
		{
			state[0] = par.state[0];
			state[1] = par.state[1];
		}
	}

	float bestValue()
	{
		float temp = Float.MAX_VALUE;
		for(int i = 1; i < Model.XMAX; i++)
		{
			for(int j = 1; j < Model.YMAX; j++)
			{
				float value = model.getTravelSpeed(i, j);
				if(value < temp)
					temp = value;
			}
		}
		return temp;
	}
	LinkedList<GameState> queuePath()
	{
		Stack<GameState> stacker = new Stack<GameState>();
		LinkedList<GameState> path = new LinkedList<GameState>();
		GameState gs = new GameState(0.0f,null,this.model);
		gs = this;
		stacker.push(gs);
		while(gs.parent != null)//stack it all up
		{
			gs = gs.parent;
			stacker.push(gs);
		}
		while(!stacker.empty())
		{
			path.add(stacker.pop());
		}
		return path;
	}

	 boolean stateEqualizer(GameState goalState)
	{
		if(state[0] == goalState.state[0] && state[1] == goalState.state[1])
			return true;
		else
			return false;
	}
	GameState roboMover(int movetype)
	{
		GameState temp = new GameState(this.cost, this, this.model);
		temp.state[0] = this.state[0];
		temp.state[1] = this.state[1];
		if(movetype == 0)
		{
			temp.state[0] += 10;
			temp.state[1] -= 10;
			if(!(temp.state[0] < 0.0f || temp.state[1] < 0.0f ||temp.state[0] > Model.XMAX || temp.state[1] > Model.YMAX))
				temp.cost += Math.sqrt(2.0d)/model.getTravelSpeed(temp.state[0],temp.state[1]);
		}
		else if(movetype == 1)
		{
			temp.state[0] += 10;
			if(!(temp.state[0] < 0.0f || temp.state[1] < 0.0f ||temp.state[0] > Model.XMAX || temp.state[1] > Model.YMAX))
				temp.cost += 1/model.getTravelSpeed(temp.state[0],temp.state[1]);
		}
		else if(movetype == 2)
		{
			temp.state[0] += 10;
			temp.state[1] += 10;
			if(!(temp.state[0] < 0.0f || temp.state[1] < 0.0f ||temp.state[0] > Model.XMAX|| temp.state[1] > Model.YMAX))
				temp.cost += Math.sqrt(2.0d)/model.getTravelSpeed(temp.state[0],temp.state[1]);
		}
		else if(movetype == 3)
		{
			temp.state[1] += 10;
			if(!(temp.state[0] < 0.0f || temp.state[1] < 0.0f ||temp.state[0] > Model.XMAX || temp.state[1] > Model.YMAX))
				temp.cost += 1/model.getTravelSpeed(temp.state[0],temp.state[1]);
		}
		else if(movetype == 4)
		{
			temp.state[1] -= 10;
			if(!(temp.state[0] < 0.0f || temp.state[1] < 0.0f ||temp.state[0] > Model.XMAX || temp.state[1] > Model.YMAX))
				temp.cost += 1/model.getTravelSpeed(temp.state[0],temp.state[1]);
		}
		else if(movetype == 5)
		{
			temp.state[0] -= 10;
			temp.state[1] -= 10;
			if(!(temp.state[0] < 0.0f || temp.state[1] < 0.0f ||temp.state[0] > Model.XMAX|| temp.state[1] > Model.YMAX))
				temp.cost += Math.sqrt(2.0d)/model.getTravelSpeed(temp.state[0],temp.state[1]);
		}
		else if(movetype == 6)
		{
			temp.state[0] -= 10;
			if(!(temp.state[0] < 0.0f || temp.state[1] < 0.0f ||temp.state[0] > Model.XMAX || temp.state[1] > Model.YMAX))
				temp.cost += 1/model.getTravelSpeed(temp.state[0],temp.state[1]);
		}
		else if(movetype == 7)
		{
			temp.state[0] -= 10;
			temp.state[1] += 10;
			if(!(temp.state[0] < 0.0f || temp.state[1] < 0.0f ||temp.state[0] > Model.XMAX|| temp.state[1] > Model.YMAX))
				temp.cost += 1/model.getTravelSpeed(temp.state[0],temp.state[1]);
		}
		if(temp.state[0] < 0.0f || temp.state[1] < 0.0f ||temp.state[0] > Model.XMAX || temp.state[1] > Model.YMAX)
		{
			return null;
		}
		else
		{
			return temp;
		}
	}
}
class beenHereComparator implements Comparator<GameState>
{
	public int compare(GameState a, GameState b)
	{
		for(int i = 0; i < 2; i++)
		{
			if(a.state[i] < b.state[i])
				return -1;
			else if(a.state[i] > b.state[i])
				return 1;
		}
		return 0;

	}
}

class aStarComparator implements Comparator<GameState>
{
	Model model;
	GameState goalState;
	aStarComparator(Model m, GameState goal)
	{
		goalState = goal;
		model = m;
	}
	public int compare(GameState a, GameState b)
	{
		double lowestCost = 1/model.getTravelSpeed(305,438);
		double distanceA = getDistance(a);
		double distanceB = getDistance(b);

		if(a.cost + a.heuristicA < b.cost + a.heuristicA)
			return -1;
		else if(a.cost + a.heuristicA > b.cost + a.heuristicA)
			return 1;
		else 
			return 0;
	}
	double getDistance(GameState a)
	{
		double distance;
		distance = Math.sqrt((a.state[0]-goalState.state[0])*(a.state[0]-goalState.state[0]) + (a.state[1] - goalState.state[1])*(a.state[1] - goalState.state[1]));
		return distance;
	}
}
public class PathPlanner
{
	beenHereComparator beenCompare;
	aStarComparator aCompare;
	TreeSet<GameState> frontier;
	TreeSet<GameState> beenHere;
	PathPlanner(Model m)
	{
		beenCompare = new beenHereComparator();
		beenHere = new TreeSet<GameState>(beenCompare);
	}
	GameState aStarSearch(GameState startState, GameState goalState, Model m)
	{
		aCompare = new aStarComparator(m, goalState);
		frontier = new TreeSet<GameState>(aCompare);
		beenHere = new TreeSet<GameState>(beenCompare);
		GameState child;
		GameState oldChild;
		startState.cost = 0.0f;
		startState.parent = null;
		goalState.parent = null;
		startState.cost = 0.0f;
		beenHere.add(startState);
		frontier.add(startState);
		while(frontier.size() > 0)
		{
			GameState test = frontier.pollFirst();
			if(test.stateEqualizer(goalState))
			{
				return test;
			}
			
			for(int i = 0; i < 8; i++)
			{
				child = test.roboMover(i);
				if(child != null)
				{
					if(beenHere.contains(child))
					{
						oldChild = beenHere.floor(child);
						if(child.cost < oldChild.cost)
						{
							oldChild.cost = child.cost;
							oldChild.parent = test;
						}
					}

					else
					{
						frontier.add(child);
						beenHere.add(child);
					}
				}
			}
			if(frontier.size() == 0)
				return test;
		}
		return new GameState(0.0f, null, m);
	}
}
}